"""Define an object to interact with the Websocket API."""
from __future__ import annotations

import asyncio
import logging
from collections.abc import Awaitable, Callable

import aiohttp

from .const import API_HOST, APP_KEY, LOGGER, STATE_CONNECTING, STATE_STOPPED, STATE_SUBSCRIBED, STATE_DISCONNECTED, EVT_DEVICE_ATTR_UPDATE
from .errors import InvalidTokenError, InvalidAuthError, WebsocketServerClosedConnectionError
from .model import Token, Device

DEFAULT_WATCHDOG_TIMEOUT = 15


class WebsocketWatchdog:
    """Define a watchdog to kick the websocket connection at intervals."""

    def __init__(
        self,
        logger: logging.Logger,
        action: Callable[..., Awaitable],
        *,
        timeout_seconds: int = DEFAULT_WATCHDOG_TIMEOUT,
    ):
        """Initialize.

        Args:
            logger: The logger to use.
            action: The coroutine function to call when the watchdog expires.
            timeout_seconds: The number of seconds before the watchdog times out.
        """
        self._action = action
        self._logger = logger
        self._loop = asyncio.get_event_loop()
        self._timeout = timeout_seconds
        self._timer_task: asyncio.TimerHandle | None = None

    def cancel(self) -> None:
        """Cancel the watchdog."""
        if self._timer_task:
            self._timer_task.cancel()
            self._timer_task = None

    async def on_expire(self) -> None:
        """Log and act when the watchdog expires."""
        self._logger.debug("Watchdog expired – calling %s",
                          self._action.__name__)
        await self._action()

    async def trigger(self) -> None:
        """Trigger the watchdog."""
        self._logger.debug(
            "Watchdog triggered – sleeping for %s seconds", self._timeout)

        if self._timer_task:
            self._timer_task.cancel()

        self._timer_task = self._loop.call_later(
            self._timeout, lambda: asyncio.create_task(self.on_expire())
        )


class VaillantWebsocketClient:  # pylint: disable=too-many-instance-attributes
    """Define the websocket client."""

    def __init__(
        self,
        token: Token,
        device: Device,
        *,
        application_key: str = APP_KEY,
        api_host: str = API_HOST,
        port: int | None = None,
        logger: logging.Logger = LOGGER,
        session: aiohttp.ClientSession | None = None,
        use_ssl: bool = True,
        verify_ssl: bool = True,
        max_retry_attemps: int | None = None,
        heartbeat_interval: int = DEFAULT_WATCHDOG_TIMEOUT,
    ) -> None:
        """Initialize.

        Args:
            application_key: Application key.
            api_version: The version of the API to query.
            logger: The logger to use.
        """

        self._session: aiohttp.ClientSession = session or aiohttp.ClientSession()
        self._use_ssl = use_ssl
        self._verify_ssl = verify_ssl
        self._token = token
        self._device = device
        self._application_key = application_key
        self._async_on_update_handler: Callable[..., Awaitable[None]] | None = None
        self._on_update_handler: Callable[..., None] | None = None
        self._logger = logger
        self._max_retry_attemps = max_retry_attemps
        self._state = None
        self._failed_attempts = 0
        self._ws_client: aiohttp.ClientWebSocketResponse | None = None
        self._watchdog = WebsocketWatchdog(logger, self.ping, timeout_seconds=heartbeat_interval)

        self._endpoint = ""
        endpoint = api_host.removeprefix("https://")
        if self._use_ssl:
            self._endpoint = f"wss://{endpoint}"
        else:
            self._endpoint = f"ws://{endpoint}"

        if port is not None:
            self._endpoint = f"{self._endpoint}:{port}"

    @property
    def state(self):
        """Return the current state."""
        return self._state

    async def connect(self) -> None:
        """Connect to websocket server."""
        
        self._state = STATE_CONNECTING

        try:
            async with self._session.ws_connect(
                f"{self._endpoint}/monitor/ws/app",
                verify_ssl=self._verify_ssl,
                headers={
                    "Authorization": f"Bearer {self._token.access_token}",
                },
            ) as ws_client:
                self._ws_client = ws_client

                await ws_client.send_json({
                    "type": "msg",
                    "productKey": self._device.product_key,
                    "mac": self._device.mac,
                    "did": self._device.id,
                })

                self._state = STATE_SUBSCRIBED

                await self._watchdog.trigger()

                async for message in ws_client:
                    if self.state == STATE_STOPPED:
                        break

                    if message.type == aiohttp.WSMsgType.TEXT:
                        msg: dict = message.json()
                        msg_type = msg.get("type", "")
                        data = msg.get("data", {})

                        self._logger.debug("Recv: %s", message)

                        if msg_type == "2":
                            device_id = msg.get("did")
                            if device_id == self._device.id:
                                self._logger.debug(
                                    "Recv atrrs for device %s => %s", device_id, data)
                                
                                if self._async_on_update_handler is not None:
                                    await self._async_on_update_handler(EVT_DEVICE_ATTR_UPDATE, { "data": data })
                                elif self._on_update_handler is not None:
                                    self._on_update_handler(EVT_DEVICE_ATTR_UPDATE, { "data": data })
                        elif msg_type == "pong":
                            self._logger.debug("Recieved pong")
                            await self._watchdog.trigger()
                        else:
                            self._logger.info("Unhandled msg: %s", msg)

                    elif message.type == aiohttp.WSMsgType.CLOSE:
                        self._logger.warning("Websocket server closing: %s, %s", message.data, message.extra)
                    elif message.type == aiohttp.WSMsgType.CLOSED:
                        self._logger.warning(
                            "Websocket connection closed")
                        break

                    elif message.type == aiohttp.WSMsgType.ERROR:
                        self._logger.error("Websocket error")
                        break

        except InvalidTokenError as error:
            self._state = STATE_STOPPED
            self._logger.warning("Websocket server auth failed: %s", error)
            raise InvalidAuthError
        except aiohttp.ClientResponseError as error:
            if error.code == 401:
                self._logger.error("Credentials rejected: %s", error)
                raise InvalidAuthError
            else:
                self._logger.error("Unexpected response received: %s", error)
            self._state = STATE_STOPPED
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError, WebsocketServerClosedConnectionError) as error:
            if self._max_retry_attemps is not None and self._failed_attempts >= self._max_retry_attemps:
                self._logger.error("Too many retry attemps. exit.")
                self._state = STATE_STOPPED
            elif self.state != STATE_STOPPED:
                retry_delay = min(2 ** (self._failed_attempts - 1) * 10, 300)
                self._failed_attempts += 1
                self._logger.error(
                    "Websocket connection failed, retrying in %ds: %s",
                    retry_delay,
                    error,
                )
                self._state = STATE_DISCONNECTED
                await asyncio.sleep(retry_delay)
        except Exception as error:  # pylint: disable=broad-except
            if self.state != STATE_STOPPED:
                self._logger.exception(
                    "Unexpected exception occurred: %s", error)
                self._state = STATE_STOPPED
        else:
            if self.state != STATE_STOPPED:
                self._state = STATE_DISCONNECTED
                await asyncio.sleep(5)

    async def listen(self):
        """Start the listening websocket."""
        self._failed_attempts = 0
        while self.state != STATE_STOPPED:
            await self.connect()

    async def close(self):
        """Close the listening websocket."""
        self._state = STATE_STOPPED
        self._watchdog.cancel()
        if self._ws_client is not None:
            await self._ws_client.close()

    async def ping(self) -> None:
        """Send ping to server."""
        if self._ws_client is not None:
            await self._ws_client.send_json({
                "type": "ping"
            })
        else:
            self._logger.error("Cannot send ping")

    def async_on_update(self, handler: Callable[..., Awaitable[None]]) -> None:
        self._async_on_update_handler = handler

    def on_update(self, handler: Callable[..., None]) -> None:
        self._on_update_handler = handler


"""Define API client to interact with the Vaillant API."""
from __future__ import annotations

import logging
from typing import Any, cast

from aiohttp import ClientSession, ClientTimeout, FormData
from aiohttp.client_exceptions import ClientError

from .const import (
    LOGGER,
    API_HOST,
    APP_KEY,
    APP_AUTH,
    DEFAULT_USER_AGENT,
)
from .errors import RequestError, InvalidAuthError, RequestError, InvalidCredentialsError
from .model import Token, Device

DEFAULT_LIMIT = 288
DEFAULT_TIMEOUT = 20


class VaillantApiClient:
    """Define the API client."""

    def __init__(
        self,
        *,
        app_key: str = APP_KEY,
        app_auth: str = APP_AUTH,
        user_agent: str = DEFAULT_USER_AGENT,
        logger: logging.Logger = LOGGER,
        session: ClientSession | None = None,
    ) -> None:
        """Initialize.

        Args:
            app_key: App key.
            app_auth: The username & password combination for authorization.
            logger: The logger to use.
            session: An optional aiohttp ClientSession.
        """
        self._application_key: str = app_key
        self._application_auth: str = app_auth
        self._access_token: str = ""
        self._user_agent: str = user_agent
        self._logger = logger
        self._session: ClientSession = session or self._new_session()

    def _new_session(self) -> ClientSession:
        return ClientSession(
            timeout=ClientTimeout(total=DEFAULT_TIMEOUT),
            raise_for_status=False,
        )

    async def _request(
        self, method: str, url: str, **kwargs: Any
    ) -> dict[str, Any]:
        """Make a request against the API.

        Args:
            method: An HTTP method.
            url: A relative API endpoint.
            **kwargs: Additional kwargs to send with the request.

        Returns:
            An API response payload.

        Raises:
            RequestError: Raised upon an underlying HTTP error.
        """

        kwargs.setdefault("params", {})
        kwargs.setdefault("headers", {})

        kwargs["headers"]["appkey"] = self._application_key

        if self._access_token != "":
            kwargs["headers"]["Authorization"] = f"Bearer {self._access_token}"

        if use_running_session := self._session and not self._session.closed:
            session = self._session
        else:
            session = self._new_session()

        try:
            async with session.request(method, f"{API_HOST}/{url}", **kwargs) as resp:
                data = {}
                if 399 < resp.status and 500 > resp.status:
                    raise InvalidAuthError
                elif 200 == resp.status:
                    data = await resp.json(content_type=None)
                else:
                    resp.raise_for_status()
        except ClientError as err:
            raise RequestError(f"Error requesting data from {url}: {err}") from err
        finally:
            if not use_running_session:
                await session.close()

        self._logger.debug("Received data for %s: %s", url, data)

        return cast(dict[str, Any], data)

    async def login(self, username: str, password: str) -> Token:
        """Login to get uid and token."""
        data = {
            "appkey": self._application_key,
            "scope": "server",
            "username": username,
            "password": password,
        }
        headers = {
            "User-Agent": self._user_agent,
            "Authorization": f"Basic {self._application_auth}",
        }
        resp = await self._request(
            "post", "auth/oauth/token?grant_type=password", data=FormData(data), headers=headers
        )
        if resp is None or resp["code"] != 200 or resp["access_token"] is None:
            raise InvalidCredentialsError

        return Token(
            app_id=self._application_key,
            username=username,
            password=password,
            access_token=resp["access_token"],
            uid=resp["user_id"],
        )

    async def get_device_list(self) -> list[Device]:
        """Get device list."""
        resp = await self._request(
            "get",
            f"app/device/getBindList?appKey={self._application_key}&version=1.0",
        )
        if resp.get("code") != 200:
            raise RequestError
        return [
            Device(
                id=d.get("did"),
                mac=d.get("mac"),
                product_key=d.get("productKey"),
                product_id=d.get("productId"),
                product_name=d.get("productName"),
                product_verbose_name=d.get("verboseName"),
                is_online=d.get("isOnline") == 1,
                is_manager=d.get("isManger") == 1,
                group_id=d.get("groupId"),
                sno=d.get("sno"),
                create_time=d.get("ctime"),
                last_offline_time=d.get("lastOfflineTime"),
                model_alias=d["modelInfo"]["aliasName"],
                model=d["modelInfo"]["model"],
                serial_number=d["serialNumber"],
                services_count=d["servicesCount"],
            )
            for d in resp["data"][0]["allBindList"]
        ]

    async def control_device(self, device_id: str, attr_name: str, value: Any):
        """Control device."""
        resp = await self._request(
            "post",
            f"app/device/control/{device_id}",
            json={
                "appKey": self._application_key,
                "data": {
                    "attrs": {
                        f"{attr_name}": value
                    }
                },
                "version": "1.0"
            }
        )
        if resp.get("code") != 200:
            raise RequestError
        